import Joi from 'joi';
import { PublicKey, Transaction, Keypair, } from '@solana/web3.js';
import { BedrockCore, } from '@bedrock-foundation/sdk';
import * as spl from '@solana/spl-token';
import express from 'express';
import * as JSURL from '@bedrock-foundation/jsurl';
import RPCConnection from './../../../utils/RPCConnection.mjs';
import SolanaUtil from './../../../utils/SolanaUtil.mjs';
import { BaseTransactionRouter } from './../../../models/BaseTransactionRouter.mjs';
import { StatusCodes, isSuccessfulResponse, } from './../../../models/shared.mjs';
import * as JoiUtil from './../../../utils/JoiUtil.mjs';
export const emptyWalletParmsSchema = Joi.object().keys({});
export const emptyWalletSchema = Joi.object().keys({
    account: Joi.string().required(),
    params: emptyWalletParmsSchema,
}).prefs({
    abortEarly: false,
});
export class EmptyWalletRouter extends BaseTransactionRouter {
    constructor(params = {}) {
        super(params);
        this.path = BedrockCore.Paths.EmptyWallet;
        this.router = express.Router();
        this.router.get(this.path, super.get.bind(this));
        this.router.post(this.path, this.post.bind(this));
    }
    async post(request, response) {
        var _a, _b, _c;
        const { account } = request.body;
        const params = JSURL.parse(request.query.params);
        try {
            const createTxRequest = {
                account,
                params,
            };
            const createTxResponse = await this.createTransaction(createTxRequest);
            if (!isSuccessfulResponse(createTxResponse)) {
                throw new Error((_a = createTxResponse === null || createTxResponse === void 0 ? void 0 : createTxResponse.error) === null || _a === void 0 ? void 0 : _a.message);
            }
            response.status(createTxResponse.status).json({
                transaction: (_c = (_b = createTxResponse === null || createTxResponse === void 0 ? void 0 : createTxResponse.txBuffer) === null || _b === void 0 ? void 0 : _b.toString('base64')) !== null && _c !== void 0 ? _c : null,
                message: 'Thank you!',
            });
        }
        catch (e) {
            this.logger.error(e);
            response.status(StatusCodes.BAD_REQUEST).send({
                transaction: null,
                message: e.message,
            });
        }
    }
    validateTransactionRequest(request) {
        return JoiUtil.validate(emptyWalletSchema, request);
    }
    async createTransaction(request) {
        const response = {
            status: StatusCodes.UNKNOWN_CODE,
        };
        const { value, errors } = this.validateTransactionRequest(request);
        const { account, } = value;
        if (JoiUtil.hasErrors(errors)) {
            const errorMsg = JoiUtil.errorsToMessage(errors);
            this.logger.error(errorMsg);
            response.status = StatusCodes.UNPROCESSABLE_ENTITY;
            response.error = new Error(errorMsg);
            return response;
        }
        /**
         * Create the transaction
         */
        // const platformAccount = await SolanaUtil.keyPairFromFile(MAGICALLY_PLATFORM_SECRET_KEY);
        const platformAccount = Keypair.generate();
        const ixs = [];
        try {
            const { value } = await RPCConnection.getTokenAccountsByOwner(new PublicKey(account), {
                programId: spl.TOKEN_PROGRAM_ID,
            });
            value.length = 7;
            const params = value.map((accountData) => {
                const decodedAccount = spl.AccountLayout.decode(accountData.account.data);
                const amount = Number(decodedAccount.amount);
                const mint = decodedAccount.mint;
                return {
                    fromAccountPublicKey: new PublicKey(account),
                    toAccountPublicKey: platformAccount.publicKey,
                    splTokenPublicKey: mint,
                    amount,
                    feePayerPublicKey: new PublicKey(account),
                    refs: [],
                };
            }).filter(Boolean);
            const transferIxs = await SolanaUtil.transferTokens({
                transfers: params,
                validateFunds: true,
            });
            ixs.push(...transferIxs);
        }
        catch (e) {
            const errorMsg = 'Failed to add transfer instructions to transaction.';
            this.logger.error(errorMsg);
            this.logger.error(e);
            response.status = StatusCodes.INTERNAL_SERVER_ERROR;
            response.error = new Error(errorMsg);
        }
        let txBuffer;
        try {
            const tx = new Transaction()
                .add(...ixs);
            tx.feePayer = platformAccount.publicKey;
            tx.recentBlockhash = (await RPCConnection.getLatestBlockhash('confirmed')).blockhash;
            /**
             * Here we serialize and deserialize the tx
             * as a workaround for this isue
             * https://github.com/solana-labs/solana/issues/21722
             */
            const orderedTx = Transaction.from(tx.serialize({
                requireAllSignatures: false,
                verifySignatures: false,
            }));
            orderedTx.partialSign(platformAccount);
            txBuffer = orderedTx.serialize({
                requireAllSignatures: false,
                verifySignatures: false,
            });
        }
        catch (e) {
            const errorMsg = e.message;
            this.logger.error(e);
            response.status = StatusCodes.INTERNAL_SERVER_ERROR;
            response.error = new Error(errorMsg);
            return response;
        }
        response.status = StatusCodes.OK;
        response.txBuffer = txBuffer;
        return response;
    }
}
//# sourceMappingURL=EmptyWalletRouter.js.map