import Joi from 'joi';
import { TransactionInstruction, PublicKey, Transaction, Keypair, } from '@solana/web3.js';
import { WaitUtil, TransactionStatuses, createNonceSocketTopic, BedrockCore, } from '@bedrock-foundation/sdk';
import jwt from 'jsonwebtoken';
import express from 'express';
import * as JSURL from '@bedrock-foundation/jsurl';
import RPCConnection from './../../utils/RPCConnection.mjs';
import { BaseTransactionRouter } from './../../models/BaseTransactionRouter.mjs';
import { StatusCodes, isSuccessfulResponse, } from './../../models/shared.mjs';
import * as JoiUtil from './../../utils/JoiUtil.mjs';
export const authorizationParmsSchema = Joi.object().keys({
    refs: Joi.array().items(Joi.string()).default([]),
});
export const authorizationSchema = Joi.object().keys({
    account: Joi.string().required(),
    params: authorizationParmsSchema,
}).prefs({
    abortEarly: false,
});
export const MEMO_PROGRAM_ID = new PublicKey('MemoSq4gqABAXKb96qnH8TysNcWxMyWCqXgDLGmfcHr');
export class AuthorizationRouter extends BaseTransactionRouter {
    constructor(params) {
        super(params);
        this.path = BedrockCore.Paths.Authorization;
        this.noncePath = BedrockCore.Paths.AuthorizationNonce;
        this.router = express.Router();
        this.redis = params.redis;
        this.io = params.io;
        this.router.get(this.path, super.get.bind(this));
        this.router.get(this.noncePath, this.createNonce.bind(this));
        this.router.post(this.path, this.post.bind(this));
    }
    async post(req, res) {
        var _a, _b, _c;
        const { account } = req.body;
        const params = JSURL.parse(req.query.params);
        try {
            const request = {
                account,
                params,
            };
            const response = await this.createTransaction(request);
            if (!isSuccessfulResponse(response)) {
                throw new Error((_a = response === null || response === void 0 ? void 0 : response.error) === null || _a === void 0 ? void 0 : _a.message);
            }
            res.status(response.status).json({
                transaction: (_c = (_b = response === null || response === void 0 ? void 0 : response.txBuffer) === null || _b === void 0 ? void 0 : _b.toString('base64')) !== null && _c !== void 0 ? _c : null,
                message: 'Authorize with Bedrock',
            });
        }
        catch (e) {
            this.logger.error(e);
            res.status(StatusCodes.BAD_REQUEST).send({
                transaction: null,
                message: e.message,
            });
        }
    }
    async createNonce(req, res) {
        var _a, _b;
        try {
            const params = JSURL.parse(req.query.params);
            const ref = (_b = (_a = params.refs) === null || _a === void 0 ? void 0 : _a[0]) !== null && _b !== void 0 ? _b : '';
            const nonce = new Keypair().publicKey.toBase58();
            this.redis.hSet('nonces', ref, nonce);
            res.status(StatusCodes.OK).json({
                nonce,
            });
        }
        catch (e) {
            this.logger.error(e);
            res.status(StatusCodes.BAD_REQUEST).send({
                nonce: null,
                message: 'Failed to create nonce.',
            });
        }
    }
    validateTransactionRequest(request) {
        return JoiUtil.validate(authorizationSchema, request);
    }
    async createTransaction(request) {
        var _a;
        const response = {
            status: StatusCodes.UNKNOWN_CODE,
        };
        const { value, errors } = this.validateTransactionRequest(request);
        const { account, params, } = value;
        if (JoiUtil.hasErrors(errors)) {
            const errorMsg = JoiUtil.errorsToMessage(errors);
            this.logger.error(errorMsg);
            response.status = StatusCodes.UNPROCESSABLE_ENTITY;
            response.error = new Error(errorMsg);
            return response;
        }
        const { refs, } = params;
        const ref = (_a = refs === null || refs === void 0 ? void 0 : refs[0]) !== null && _a !== void 0 ? _a : '';
        const nonce = await this.redis.hGet('nonces', ref);
        if (!nonce) {
            const errorMsg = 'Invalid request. Nonce not found.';
            this.logger.error(errorMsg);
            response.status = StatusCodes.BAD_REQUEST;
            response.error = new Error(errorMsg);
            return response;
        }
        /**
         * Broadcast the scan to the client
         */
        const nonceSocketTopic = createNonceSocketTopic(nonce);
        const customerPublicKey = new PublicKey(account);
        const wallet = customerPublicKey.toBase58();
        const authorizationData = {
            wallet,
            status: TransactionStatuses.Scanned,
            signature: null,
            token: null,
        };
        this.io.emit(nonceSocketTopic, authorizationData);
        /**
         * Create the transaction
         */
        let txBuffer;
        const keypair = Keypair.generate();
        try {
            const memoIx = new TransactionInstruction({
                programId: MEMO_PROGRAM_ID,
                keys: [
                    {
                        pubkey: customerPublicKey,
                        isSigner: true,
                        isWritable: false,
                    },
                    {
                        pubkey: new PublicKey(keypair.publicKey),
                        isSigner: true,
                        isWritable: false,
                    },
                ],
                data: Buffer.from('BEDROCK_AUTHORIZATION', 'utf8'),
            });
            const tx = new Transaction().add(memoIx);
            tx.feePayer = customerPublicKey;
            tx.recentBlockhash = (await RPCConnection.getLatestBlockhash('confirmed')).blockhash;
            /**
             * Here we serialize and deserialize the tx
             * as a workaround for this isue
             * https://github.com/solana-labs/solana/issues/21722
             */
            const orderedTx = Transaction.from(tx.serialize({
                requireAllSignatures: false,
                verifySignatures: false,
            }));
            orderedTx.partialSign(keypair);
            txBuffer = orderedTx.serialize({
                requireAllSignatures: false,
                verifySignatures: false,
            });
            const createJwt = (signature, wallet) => {
                return jwt.sign({
                    signature,
                    wallet,
                }, 'BEDROCK_AUTHORIZATION_SECRET');
            };
            const confirmTransaction = async () => {
                try {
                    // try to confirm the transaction up to 30 times
                    for (let i = 0; i < 30; i++) {
                        try {
                            await WaitUtil.wait(2000);
                            const signatures = await RPCConnection.getSignaturesForAddress(keypair.publicKey, {}, 'confirmed');
                            if (signatures.length > 0) {
                                const signature = signatures[0].signature;
                                const authorizationData = {
                                    wallet,
                                    status: TransactionStatuses.Confirmed,
                                    signature,
                                    token: createJwt(signature, wallet),
                                };
                                this.io.emit(nonceSocketTopic, authorizationData);
                                break;
                            }
                        }
                        catch (e) {
                            this.logger.log(e);
                            this.logger.log(`Failed to confirm tranaction with ref ${ref}`);
                        }
                    }
                }
                catch (e) {
                    this.logger.error('Failed to confirm transaction');
                    this.logger.error(e);
                }
            };
            confirmTransaction();
        }
        catch (e) {
            const errorMsg = e.message;
            this.logger.error(e);
            response.status = StatusCodes.INTERNAL_SERVER_ERROR;
            response.error = new Error(errorMsg);
            return response;
        }
        response.status = StatusCodes.OK;
        response.txBuffer = txBuffer;
        return response;
    }
}
//# sourceMappingURL=AuthorizationRouter.js.map