import Joi from 'joi';
import { LAMPORTS_PER_SOL, SystemProgram, PublicKey, Transaction, } from '@solana/web3.js';
import { TokenTypes, TokenInfo, BedrockCore, } from '@bedrock-foundation/sdk';
import express from 'express';
import * as JSURL from '@bedrock-foundation/jsurl';
import RPCConnection from './../../utils/RPCConnection.mjs';
import SolanaUtil from './../../utils/SolanaUtil.mjs';
import { BaseTransactionRouter } from './../../models/BaseTransactionRouter.mjs';
import { StatusCodes, isSuccessfulResponse, } from './../../models/shared.mjs';
import * as JoiUtil from './../../utils/JoiUtil.mjs';
export const transferParamsSchema = Joi.object().keys({
    wallet: Joi.string().required(),
    token: Joi.string().required(),
    quantity: Joi.number().optional(),
    size: Joi.number().optional(),
    icon: Joi.string().optional(),
    label: Joi.string().optional(),
    refs: Joi.array().items(Joi.string()).default([]),
});
export const transferSchema = Joi.object().keys({
    account: Joi.string().required(),
    params: transferParamsSchema,
}).prefs({
    abortEarly: false,
});
export class TransferRouter extends BaseTransactionRouter {
    constructor(params = {}) {
        super(params);
        this.path = BedrockCore.Paths.Transfer;
        this.router = express.Router();
        this.router.get(this.path, super.get.bind(this));
        this.router.post(this.path, this.post.bind(this));
    }
    async post(req, res) {
        var _a, _b, _c;
        const { account } = req.body;
        const params = JSURL.parse(req.query.params);
        try {
            const request = {
                account,
                params,
            };
            const response = await this.createTransaction(request);
            if (!isSuccessfulResponse(response)) {
                throw new Error((_a = response === null || response === void 0 ? void 0 : response.error) === null || _a === void 0 ? void 0 : _a.message);
            }
            res.status(response.status).json({
                transaction: (_c = (_b = response === null || response === void 0 ? void 0 : response.txBuffer) === null || _b === void 0 ? void 0 : _b.toString('base64')) !== null && _c !== void 0 ? _c : null,
                message: 'Thank you!',
            });
        }
        catch (e) {
            this.logger.error(e);
            res.status(StatusCodes.BAD_REQUEST).send({
                transaction: null,
                message: e.message,
            });
        }
    }
    validateTransactionRequest(request) {
        return JoiUtil.validate(transferSchema, request);
    }
    async createTransaction(request) {
        var _a, _b, _c;
        const response = {
            status: StatusCodes.UNKNOWN_CODE,
        };
        const { value, errors } = this.validateTransactionRequest(request);
        const { account, params, } = value;
        if (JoiUtil.hasErrors(errors)) {
            const errorMsg = JoiUtil.errorsToMessage(errors);
            this.logger.error(errorMsg);
            response.status = StatusCodes.UNPROCESSABLE_ENTITY;
            response.error = new Error(errorMsg);
            return response;
        }
        const { wallet, token, quantity, size, refs, } = params;
        /**
         * Create the transaction
         */
        const customerPublicKey = new PublicKey(account);
        const merchantPublicKey = new PublicKey(wallet);
        const ixs = [];
        try {
            if (token === TokenTypes.SOL) {
                const amount = (_a = (size ? size * LAMPORTS_PER_SOL : quantity)) !== null && _a !== void 0 ? _a : 0;
                const merchantTransferIx = SystemProgram.transfer({
                    fromPubkey: customerPublicKey,
                    toPubkey: merchantPublicKey,
                    lamports: amount,
                });
                refs === null || refs === void 0 ? void 0 : refs.forEach((ref) => {
                    merchantTransferIx.keys.push({ pubkey: new PublicKey(ref), isWritable: false, isSigner: false });
                });
                ixs.push(merchantTransferIx);
            }
            else if (token === TokenTypes.USDC) {
                const usdcTokenInfo = TokenInfo[TokenTypes.USDC];
                const usdcTokenAddress = new PublicKey(usdcTokenInfo.address);
                const amount = (_b = (size ? size * (10 ** usdcTokenInfo.decimals) : quantity)) !== null && _b !== void 0 ? _b : 0;
                const params = [
                    {
                        fromAccountPublicKey: customerPublicKey,
                        toAccountPublicKey: merchantPublicKey,
                        splTokenPublicKey: usdcTokenAddress,
                        amount,
                        feePayerPublicKey: customerPublicKey,
                        refs: (_c = refs === null || refs === void 0 ? void 0 : refs.map((ref) => new PublicKey(ref))) !== null && _c !== void 0 ? _c : [],
                    },
                ];
                const transferIxs = await SolanaUtil.transferTokens({
                    transfers: params,
                    validateFunds: true,
                });
                ixs.push(...transferIxs);
            }
        }
        catch (e) {
            const errorMsg = 'Failed to add transfer instructions to transaction.';
            this.logger.error(errorMsg);
            this.logger.error(e);
            response.status = StatusCodes.INTERNAL_SERVER_ERROR;
            response.error = new Error(errorMsg);
            return response;
        }
        let txBuffer;
        try {
            const tx = new Transaction()
                .add(...ixs);
            tx.feePayer = customerPublicKey;
            tx.recentBlockhash = (await RPCConnection.getLatestBlockhash('confirmed')).blockhash;
            /**
             * Here we serialize and deserialize the tx
             * as a workaround for this isue
             * https://github.com/solana-labs/solana/issues/21722
             */
            const orderedTx = Transaction.from(tx.serialize({
                requireAllSignatures: false,
                verifySignatures: false,
            }));
            txBuffer = orderedTx.serialize({
                requireAllSignatures: false,
                verifySignatures: false,
            });
        }
        catch (e) {
            const errorMsg = e.message;
            this.logger.error(e);
            response.status = StatusCodes.INTERNAL_SERVER_ERROR;
            response.error = new Error(errorMsg);
            return response;
        }
        response.status = StatusCodes.OK;
        response.txBuffer = txBuffer;
        return response;
    }
}
//# sourceMappingURL=TransferRouter.js.map